"""
Implementation of the solution proposed on the README.MD of https://github.com/YoannQDQ/qgis-layer-tree-icons
Monkey patch the default menuProvider to add new action especially to Analysis node where regression models results
can be accessed through the contextual menu
"""
from enum import Enum
from typing import Optional, Callable, List

from PyQt5.QtCore import QCoreApplication
from qgis.PyQt.QtCore import QObject, QEvent
from qgis.PyQt.QtWidgets import QAction, QMenu
from qgis._core import QgsLayerTree, QgsVectorLayer, QgsRasterLayer
from qgis._gui import QgisInterface, QgsLayerTreeView, QgsLayerTreeViewMenuProvider

from landsklim.lisdqs.lisdqs_analysis import LISDQSAnalysis
from landsklim.lisdqs.lisdqs_interpolation import LISDQSInterpolation, LISDQSInterpolationType


class LayerTreeContextMenuManager(QObject):

    def __init__(self, iface: QgisInterface, parent=None):
        super().__init__(parent)
        self._iface: QgisInterface = iface
        self.view: QgsLayerTreeView = iface.layerTreeView()
        self.menuProvider: QgsLayerTreeViewMenuProvider = self.view.menuProvider()

        self.providers: List[Callable] = []

        self.patch()

        self.view.viewport().installEventFilter(self)

    # Made event argument optional because mocked MenuProvider for tests doesn't pass an event parameter to this function (why ?)
    def createContextMenu(self, event: QEvent=None):
        menu: QMenu = self._iface.layerTreeView().menuProvider()._original()
        for provider in self._iface.layerTreeView().menuProvider().providers:  # type: Callable
            try:
                provider(menu, event)
            except TypeError:
                provider(menu)
        return menu

    def patch(self):

        if not hasattr(self.menuProvider, "_original"):
            self.menuProvider._original = self.menuProvider.createContextMenu
        if not hasattr(self.menuProvider, "providers"):
            self.menuProvider.providers = []

        self.menuProvider.createContextMenu = self.createContextMenu  # = lambda event: print(event) and self.createContextMenu(event)

    def eventFilter(self, obj: QObject, event: QEvent):
        if event.type() == QEvent.ContextMenu:
            menu: QMenu = self.menuProvider.createContextMenu(event)
            menu.exec(self.view.mapToGlobal(event.pos()))
            return True
        return False

    def addProvider(self, provider: Callable):
        if not callable(provider):
            return
        if provider in self.menuProvider.providers:
            return
        self.providers.append(provider)
        self.menuProvider.providers.append(provider)

    def removeProvider(self, provider: Callable):
        try:
            self.menuProvider.providers.remove(provider)
        except ValueError:
            pass

    def __del__(self):
        for provider in self.providers:
            self.removeProvider(provider)
        self.view.viewport().removeEventFilter(self)


class LISDQSAnalysisChartType(Enum):
    Correlation = 0
    AnalysisSeriesByStation = 1
    AnalysisSeriesByPhase = 2
    AnalysisExplicativeVariableSeries = 3
    AnalysisExplicativeVariableDistribution = 4
    AnalysisExplicativeVariable = 5


class LISDQSMenuProvider:
    def __init__(self, iface: QgisInterface, lisdqs_instance: "LISDQS"):
        self.view: QgsLayerTreeView = iface.layerTreeView()
        self.__lisdqs_instance: "LISDQS" = lisdqs_instance
        self.__handle_on_analysis_situation_click: Optional[Callable[[LISDQSAnalysis, int], None]] = None
        self.__handle_on_analysis_chart_click: Optional[Callable[[LISDQSAnalysis, LISDQSAnalysisChartType], None]] = None
        self.__handle_on_analysis_delete: Optional[Callable[[LISDQSAnalysis], None]] = None
        self.__handle_on_interpolation_delete: Optional[Callable[[LISDQSAnalysis, LISDQSInterpolation], None]] = None
        self.__handle_on_save_as_netcdf_click: Optional[Callable[[str], None]] = None
        self.__handle_on_interpolation_save_as_netcdf_click: Optional[Callable[[LISDQSAnalysis, LISDQSInterpolation, LISDQSInterpolationType], None]] = None

    def tr(self, string: str) -> str:
        res = QCoreApplication.translate('LISDQSMenuProvider', string)
        return res

    def handle_on_analysis_situation_click(self, handle: Callable[[LISDQSAnalysis, int], None]):
        self.__handle_on_analysis_situation_click = handle

    def handle_on_analysis_chart_click(self, handle: Callable[[LISDQSAnalysis, LISDQSAnalysisChartType], None]):
        self.__handle_on_analysis_chart_click = handle

    def handle_on_save_as_netcdf_click(self, handle: Callable[[str], None]):
        self.__handle_on_save_as_netcdf_click = handle

    def handle_on_interpolation_save_as_netcdf_click(self, handle: Callable[[LISDQSAnalysis, LISDQSInterpolation, LISDQSInterpolationType], None]):
        self.__handle_on_interpolation_save_as_netcdf_click = handle

    def handle_on_analysis_delete_click(self, handle: Callable[[LISDQSAnalysis], None]):
        self.__handle_on_analysis_delete = handle

    def handle_on_interpolation_delete_click(self, handle: Callable[[LISDQSAnalysis, LISDQSInterpolation], None]):
        self.__handle_on_interpolation_delete = handle

    def get_analysis_from_property(self, lisdqs_object_property: str) -> LISDQSAnalysis:
        """
        Retrieve LISDQSAnalysis from lisdqs_object node property
        """
        index_configuration: int = int(lisdqs_object_property.replace("configuration_", "").split("_analysis_")[0])
        index_analysis: int = int(lisdqs_object_property.replace("configuration_{0}_analysis_".format(index_configuration), "").split("_interpolation_")[0])
        return self.__lisdqs_instance.get_lisdqs_project().get_configurations()[index_configuration].get_analysis()[index_analysis]

    def get_interpolation_from_property(self, lisdqs_object_property: str) -> LISDQSInterpolation:
        index_configuration: int = int(lisdqs_object_property.replace("configuration_", "").split("_analysis_")[0])
        index_analysis: int = int(lisdqs_object_property.replace("configuration_{0}_analysis_".format(index_configuration), "").split("_interpolation_")[0])
        index_interpolations: int = int(lisdqs_object_property.replace("configuration_{0}_analysis_{1}_interpolation_".format(index_configuration, index_analysis), ""))
        return self.__lisdqs_instance.get_lisdqs_project().get_configurations()[index_configuration].get_analysis()[index_analysis].get_interpolations()[index_interpolations]

    def __call__(self, menu):
        lisdqs_object: str = self.view.currentGroupNode().customProperty("lisdqs_object", "")

        if not QgsLayerTree.isLayer(self.view.currentNode()) and "configuration_" in lisdqs_object and "analysis_" in lisdqs_object and "interpolation_" not in lisdqs_object:
            menu.addSeparator()
            analysis: LISDQSAnalysis = self.get_analysis_from_property(lisdqs_object)
            """lb_correlation = lambda _, arg_analysis=analysis: self.__handle_on_analysis_chart_click(arg_analysis, LISDQSAnalysisChartType.Correlation)
            action_correlation: QAction = QAction("Tableau de corrélation", menu)
            action_correlation.triggered.connect(lb_correlation)
            menu.addAction(action_correlation)
            menu.addSeparator()

            lb_series_by_station = lambda _, arg_analysis=analysis: self.__handle_on_analysis_chart_click(arg_analysis, LISDQSAnalysisChartType.AnalysisSeriesByStation)
            action_series_by_station: QAction = QAction("Stat série par station", menu)
            action_series_by_station.triggered.connect(lb_series_by_station)
            menu.addAction(action_series_by_station)
            lb_series_by_phase = lambda _, arg_analysis=analysis: self.__handle_on_analysis_chart_click(arg_analysis, LISDQSAnalysisChartType.AnalysisSeriesByPhase)
            action_series_by_phase: QAction = QAction("Stat série par phase", menu)
            action_series_by_phase.triggered.connect(lb_series_by_phase)
            menu.addAction(action_series_by_phase)
            if not analysis.is_local():
                lb_ev_distribution = lambda _, arg_analysis=analysis: self.__handle_on_analysis_chart_click(arg_analysis, LISDQSAnalysisChartType.AnalysisExplicativeVariableDistribution)
                action_ev_distribution: QAction = QAction("Stat distrib. var. exp.", menu)
                action_ev_distribution.triggered.connect(lb_ev_distribution)
                menu.addAction(action_ev_distribution)
                lb_ev_series = lambda _, arg_analysis=analysis: self.__handle_on_analysis_chart_click(arg_analysis, LISDQSAnalysisChartType.AnalysisExplicativeVariableSeries)
                action_ev_series: QAction = QAction("Stat série var. exp.", menu)
                action_ev_series.triggered.connect(lb_ev_series)
                menu.addAction(action_ev_series)
                lb_ev = lambda _, arg_analysis=analysis: self.__handle_on_analysis_chart_click(arg_analysis, LISDQSAnalysisChartType.AnalysisExplicativeVariable)
                action_ev: QAction = QAction("Stat var. exp.", menu)
                action_ev.triggered.connect(lb_ev)
                menu.addAction(action_ev)

            menu.addSeparator()
            for i, situation in enumerate(analysis.get_situations_names()):  # type: str
                action_regression = QAction(situation, menu)
                # I presume QAction populate the first parameter of the lambda by a bool value, this param is ignored
                lb = lambda _, arg_analysis=analysis, arg_i=i: self.__handle_on_analysis_situation_click(arg_analysis, arg_i)
                action_regression.triggered.connect(lb)
                menu.addAction(action_regression)"""

            menu.addSeparator()
            lb_ev_delete = lambda _, arg_analysis=analysis: self.__handle_on_analysis_delete(arg_analysis)
            action_ev_delete: QAction = QAction(self.tr("Delete analysis"), menu)
            action_ev_delete.triggered.connect(lb_ev_delete)
            menu.addAction(action_ev_delete)

        if not QgsLayerTree.isLayer(self.view.currentNode()) and "configuration_" in lisdqs_object and "analysis_" in lisdqs_object and "interpolation_" in lisdqs_object:
            analysis: LISDQSAnalysis = self.get_analysis_from_property(lisdqs_object)
            interpolation: LISDQSInterpolation = self.get_interpolation_from_property(lisdqs_object)
            menu.addSeparator()
            lb_ev_delete = lambda _, arg_analysis=analysis, arg_interpolation=interpolation: self.__handle_on_interpolation_delete(arg_analysis, arg_interpolation)
            action_ev_delete: QAction = QAction(self.tr("Delete interpolation"), menu)
            action_ev_delete.triggered.connect(lb_ev_delete)
            menu.addAction(action_ev_delete)
            if interpolation.is_on_grid():
                menu.addSeparator()
                for phase_type in interpolation.get_interpolation_types():  # type: LISDQSInterpolationType
                    lb_ev_export = lambda _, arg_analysis=analysis, arg_interpolation=interpolation, arg_phase_type=phase_type: self.__handle_on_interpolation_save_as_netcdf_click(arg_analysis, arg_interpolation, arg_phase_type)
                    action_ev_export: QAction = QAction(self.tr("[{0}] Export as NetCDF").format(phase_type.str()), menu)
                    action_ev_export.setData('export_netcdf')
                    action_ev_export.triggered.connect(lb_ev_export)
                    menu.addAction(action_ev_export)

        """if isinstance(self.view.currentLayer(), QgsRasterLayer):
            menu.addSeparator()
            action_netcdf: QAction = QAction(self.tr("Export as NetCDF"), menu)
            action_netcdf.setData('export_netcdf')
            lb = lambda _, arg_source=self.view.currentLayer().source(): self.__handle_on_save_as_netcdf_click(arg_source)
            action_netcdf.triggered.connect(lb)
            menu.addAction(action_netcdf)"""